// Elan no Nanika plugin for AmaRecCo
// Author: elz (http://zoome.jp/elz/)
// Released under the Artistic License

#define PLUGIN_VERSION "0.1"

#include <stdio.h>
#include <windows.h>
#include <windowsx.h>
#include <wingdi.h>
#include <time.h>
#include "PluginInput.h"
#include "resource.h"
#include "config.h"

// Amarecoco Defined Methods
//---------------------------------------------------------------------
BOOL initPlugin(void);
BOOL exitPlugin(void);
BOOL dispConfigDialog(HWND hParent, int clas);
BOOL presetRecord(S_INPUT_PLUGIN_CAPTUR *psipc);
BOOL startRecord(S_INPUT_PLUGIN_CAPTUR *psipc);
BOOL stopRecord(S_INPUT_PLUGIN_CAPTUR *psipc);
BOOL captureMain(S_INPUT_PLUGIN_CAPTUR *psipc);


// User Defined Methods
//---------------------------------------------------------------------
extern "C" BOOL WINAPI DllMain(HINSTANCE hInstDLL, DWORD dwReason, LPVOID lpReserved);

BOOL CALLBACK procMainDialog(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
BOOL loadSettingToDialog(HWND hWnd);
BOOL insertDateFormatToEditCtrl(HWND hWnd);
BOOL updateSettingMainDialog(HWND hWnd);
void updatePreview(HWND hWnd);
void setFont(HDC hdc);
void paintPreviewBackground(HDC hdc, const RECT& rc);
inline void drawTimeStamp(HDC hdc, int iw, int ih);

BOOL CALLBACK procRecordDialog(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
BOOL updateSettingRecordDialog(HWND hWnd);
BOOL moveRecordDialog(HWND hWnd);
BOOL closeRecordDialog(HWND hWnd);


// Global Variables
//---------------------------------------------------------------------
HINSTANCE    hInst;
HDC          hDC;
HWND         hRecordDialog;
HWND         hMainDialog;
bool         canEnableTimeStamp;
bool         canContinueRecording;
ConfigValues config;

INPUT_PLUGIN_TABLE input_plugin_table = {
    sizeof(INPUT_PLUGIN_TABLE),
    INPUT_PLUGIN_TABLE_VERSION,
    PLUGIN_NAME,
    "",
    initPlugin,
    exitPlugin,
    presetRecord,
    startRecord,
    stopRecord,
    NULL,
    captureMain,
    NULL,
    dispConfigDialog,
    NULL,
};


//---------------------------------------------------------------------
extern "C" INPUT_PLUGIN_TABLE __declspec(dllexport) * __stdcall
    GetInputPluginTable(void)
{
    return &input_plugin_table;
}

extern "C" BOOL WINAPI
    DllMain(HINSTANCE hInstDLL, DWORD dwReason, LPVOID lpReserved)
{
    switch (dwReason) {
    case DLL_PROCESS_ATTACH:
        ::hInst = hInstDLL;
        break;

    case DLL_PROCESS_DETACH:
        break;
    }
    return TRUE;
}

//---------------------------------------------------------------------
BOOL
initPlugin(void)
{
    ::hDC = NULL;
    config.setPlugin(&input_plugin_table);

    return TRUE;
}

BOOL
exitPlugin(void)
{
    /*
    MessageBox(NULL, "exitPlugin", "OK", MB_OK);
    ::hDC = NULL;
    ::hRecordDialog = NULL;
    ::hMainDialog = NULL;
    */

    return TRUE;
}

BOOL
presetRecord(S_INPUT_PLUGIN_CAPTUR *psipc)
{
    config.setIndex(psipc->clas);
    config.reloadAll();

    if (psipc->flag & IPC_FLAG_TIMESTAMP) {
        psipc->flag = psipc->flag & ~IPC_FLAG_TIMESTAMP;
        ::canEnableTimeStamp = true;
    }
    else {
        ::canEnableTimeStamp = false;
    }

    return TRUE;
}

BOOL
startRecord(S_INPUT_PLUGIN_CAPTUR *psipc)
{
    ::hDC = GetDC(NULL);
    ::canContinueRecording = true;

    if (::canEnableTimeStamp) {
        setFont(psipc->dc);
    }

    if (!::hRecordDialog && config.buttonMode) {
        ::hRecordDialog = CreateDialog(::hInst, MAKEINTRESOURCE(IDD_REC), NULL,
                                       (DLGPROC)procRecordDialog);
    }

    return TRUE;
}

BOOL
dispConfigDialog(HWND hParent, int clas)
{
    if (::hMainDialog) {
        PostMessage(::hMainDialog, WM_CLOSE, 0, 0);
    }
    else {
        config.setIndex(clas);
        ::hMainDialog = CreateDialog(::hInst, MAKEINTRESOURCE(IDD_CONFIG), hParent, (DLGPROC)procMainDialog);
    }

    return TRUE;
}

BOOL
stopRecord(S_INPUT_PLUGIN_CAPTUR *psipc)
{
    ::canContinueRecording = 0;

    if (::hDC) {
        HGDIOBJ h;

        if (h = GetCurrentObject(::hDC, OBJ_FONT)) {
            DeleteObject(h);
        }

        if (h = GetCurrentObject(::hDC, OBJ_BRUSH)) {
            DeleteObject(h);
        }

        ReleaseDC(NULL, ::hDC);
        ::hDC = NULL;
    }

    if (::hRecordDialog) {
        PostMessage(::hRecordDialog, WM_CLOSE, 0, 0);
    }

    return TRUE;
}

BOOL
captureMain(S_INPUT_PLUGIN_CAPTUR *psipc)
{
    if (!::canContinueRecording) {
        return FALSE;
    }

    DWORD flag = SRCCOPY;

    if (psipc->flag & IPC_FLAG_LAYERED) {
        flag |= CAPTUREBLT;
    }

    if (BitBlt(psipc->dc, 0, 0, psipc->w, psipc->h, ::hDC, psipc->x, psipc->y, flag) == 0) {
        return FALSE;
    }

    if (::canEnableTimeStamp) {
        drawTimeStamp(psipc->dc, psipc->bmp_w, psipc->bmp_h);
    }

    return TRUE;
}

//---------------------------------------------------------------------
BOOL CALLBACK
procMainDialog(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
    case WM_INITDIALOG:
        loadSettingToDialog(hWnd);
        break;

    case WM_PAINT:
        updatePreview(hWnd);
        break;

    case WM_COMMAND:
        switch (LOWORD(wParam)) {
        case IDC_BT_PREVIEW:
            SendMessage(hWnd, WM_PAINT, 0, TRUE);
            break;

        case IDC_BT_DEFAULT:
            SetDlgItemText(hWnd, IDC_EDIT1, DEF_TIMESTAMPFORMAT);
            break;

        case IDC_BT_INSERT:
            insertDateFormatToEditCtrl(hWnd);
            break;

        case IDC_COMBO1:
            if (HIWORD(wParam) == CBN_SELCHANGE) {
                SetFocus(GetDlgItem(hWnd, IDC_EDIT1));
            }
            break;

        case IDC_BT_OK:
            updateSettingMainDialog(hWnd);
            SendMessage(hWnd, WM_CLOSE, 0, TRUE);
            break;

        case IDC_BT_CANCEL:
            SendMessage(hWnd, WM_CLOSE, 0, FALSE);
            break;
        }
        break;

    case WM_CLOSE:
        DestroyWindow(hWnd);
        ::hMainDialog = NULL;
        return TRUE;
    }

    return FALSE;
}

BOOL
loadSettingToDialog(HWND hWnd)
{
    config.reloadAll();

    SetDlgItemText(hWnd, IDC_EDIT1, config.tsFormat);

    switch (config.buttonMode) {
    case DISABLE:
        Button_SetCheck(GetDlgItem(hWnd, IDC_RADIO1), BST_CHECKED);
        break;
    case PREVIOUS_POSITION:
        Button_SetCheck(GetDlgItem(hWnd, IDC_RADIO2), BST_CHECKED);
        break;
    case CORNER:
        Button_SetCheck(GetDlgItem(hWnd, IDC_RADIO3), BST_CHECKED);
        break;
    }

    int i;
    char buf[32];
    HWND hList = GetDlgItem(hWnd, IDC_COMBO1);

    for (i = 0; i <= 16; i++) {
        LoadString(::hInst, IDS_STRING0 + i, (LPSTR)&buf, sizeof(buf) - 1);
        SendMessage(hList, CB_ADDSTRING, 0, (LPARAM)buf);
    }

    SendMessage(hList, CB_SETCURSEL, 0, 0);

    HWND hEdit = GetDlgItem(hWnd, IDC_EDIT1);
    SetFocus(hEdit);
    SendMessage(hEdit, EM_SETSEL, SendMessage(hEdit, WM_GETTEXTLENGTH, 0, 0), -1);

    return TRUE;
}

BOOL
insertDateFormatToEditCtrl(HWND hWnd)
{
    int line = 0;
    char buf[32];
    HWND hList = GetDlgItem(hWnd, IDC_COMBO1);
    HWND hEdit = GetDlgItem(hWnd, IDC_EDIT1);

    line = SendMessage(hList, CB_GETCURSEL, 0, 0);
    LoadString(::hInst, IDS_STRING50 + line, (LPSTR)buf, sizeof(buf) - 1);
    SendMessage(hEdit, EM_REPLACESEL, 1, (LPARAM)buf);
    SetFocus(hEdit);

    return TRUE;
}

BOOL updateSettingMainDialog(HWND hWnd)
{
    if (Button_GetCheck(GetDlgItem(hWnd, IDC_RADIO1))) {
        config.buttonMode = DISABLE;
    }
    else if (Button_GetCheck(GetDlgItem(hWnd, IDC_RADIO2))) {
        config.buttonMode = PREVIOUS_POSITION;
    }
    else if (Button_GetCheck(GetDlgItem(hWnd, IDC_RADIO3))) {
        config.buttonMode = CORNER;
    }

    char format[MAX_LEN + 1];
    GetDlgItemText(hWnd, IDC_EDIT1, format, MAX_LEN);
    strncpy(config.tsFormat, format, MAX_LEN);

    config.updateAll();

    return TRUE;
}

void
updatePreview(HWND hWnd)
{
    config.reloadAll();

    HWND hCtrl = NULL;
    HRGN hClip = NULL;
    HDC hImage = NULL;
    RECT rc;

    hCtrl = GetDlgItem(hWnd, IDC_IMAGE);
    GetClientRect(hCtrl, &rc);
    hImage = GetDC(hCtrl);

    setFont(hImage);

    hClip = CreateRectRgn(0, 0, rc.right, rc.bottom);
    SelectClipRgn(hImage, hClip);
    DeleteObject(hClip);

    paintPreviewBackground(hImage, rc);
    GetDlgItemText(hWnd, IDC_EDIT1, config.tsFormat, MAX_LEN);
    drawTimeStamp(hImage, rc.right, rc.bottom);

    // clean
    DeleteObject(GetCurrentObject(hImage, OBJ_FONT));
    DeleteObject(GetCurrentObject(hImage, OBJ_BRUSH));
    ReleaseDC(NULL, hImage);

    // disp title
    char caption[MAX_LEN + 1];
    char setting[MAX_LEN + 1];
    LoadString(::hInst, IDS_STRING_SETTING, (LPSTR)&setting, MAX_LEN);
    snprintf(caption, MAX_LEN, "%s %s [%s %d]", config.pluginName, PLUGIN_VERSION, setting, config.index + 1);
    SetWindowText(hWnd, caption);
}

void
setFont(HDC hdc)
{
    SetTextColor(hdc, config.fontColor);
    SelectObject(hdc, CreateSolidBrush(config.backgroundColor));
    SetBkMode(hdc, TRANSPARENT);

    SelectObject(hdc, CreateFont(
                     config.fontHeight,
                     0,
                     0,
                     0,
                     FW_MEDIUM,
                     FALSE,
                     FALSE,
                     FALSE,
                     DEFAULT_CHARSET,
                     OUT_RASTER_PRECIS,
                     CLIP_DEFAULT_PRECIS,
                     ANTIALIASED_QUALITY,
                     DEFAULT_PITCH || FF_MODERN,
                     config.fontName
                 ));

    return;
}

void
paintPreviewBackground(HDC hdc, const RECT& rc)
{
    int color;
    HBRUSH br = NULL;

    srand((unsigned)time(NULL));
    color = rand() % 5;

    switch (color) {
    case 0: // blue green
        br = CreateSolidBrush(RGB(50, 148, 189));
        break;

    case 1: // yellow
        br = CreateSolidBrush(RGB(190, 150, 39));
        break;

    case 2: // green
        br = CreateSolidBrush(RGB(90, 175, 51));
        break;

    case 3: // purple
        br = CreateSolidBrush(RGB(140, 80, 120));
        break;

    default: // red
        br = CreateSolidBrush(RGB(200, 120, 80));
        break;
    }

    FillRect(hdc, &rc, br);
    DeleteObject(br);

    RECT sqrc;
    int x, y, w, h, r, g, b, sub;

    w = rc.right / 14;
    h = w;

    for (int i = 0; i < 14 * (rc.bottom / h + 1); i++) {
        switch (color) {
        case 0: // blue
            r = 50 + rand() % 25;
            g = r + 50 + rand() % 35;
            b = g + 65 + rand() % 45;
            if (b > 255) b = 255;
            break;

        case 1: // yellow
            b = 100 + rand() % 25;
            g = b + 40 + rand() % 45;
            r = g + 30 + rand() % 50 - 15;
            if (r > 255) r = 255;
            break;

        case 2: // green
            r = 90 + rand() % 25;
            b = 60 + rand() % 40;
            g = b + 90 + rand() % 50;
            if (g > 255) g = 255;
            break;

        case 3: // purple
            g = 80 + rand() % 25;
            b = g + 40 + rand() % 35;
            r = b + 20 + rand() % 45 - 15;
            if (r > 255) r = 255;
            break;

        default: // red
            b = 80 + rand() % 25;
            g = b + 10 + rand() % 35;
            r = g + 65 + rand() % 45;
            if (r > 255) r = 255;
            break;
        }

        x = (i % 14) * w + 3;
        y = (i / 14) * h;
        sub = rand() % 8;
        SetRect(&sqrc, x + sub / 2, y + sub / 2, x + w - sub, y + h - sub);

        br = CreateSolidBrush(RGB(r, g, b));
        FillRect(hdc, &sqrc, br);
        DeleteObject(br);
    }

    return;
}

inline void
drawTimeStamp(HDC hdc, int iw, int ih)
{
    static char timestamp[MAX_LEN + 1];
    static SIZE size;
    static time_t now;
    int x, y, w, h;

    time(&now);
    strftime(timestamp, MAX_LEN, config.tsFormat, localtime(&now));

    GetTextExtentPoint32(hdc, timestamp, strlen(timestamp), &size);

    w = size.cx + MERGIN_X * 2;
    h = size.cy + MERGIN_Y * 2;

    switch (config.fontAlign) {
    case TOP_RIGHT:
        x = iw - w;
        y = 0;
        break;

    case BOTTOM_LEFT:
        x = 0;
        y = ih - h;
        break;

    case BOTTOM_RIGHT:
        x = iw - w;
        y = ih - h;
        break;

    default: // TOP_LEFT
        x = 0;
        y = 0;
    }

    if (!config.isTransparent) PatBlt(hdc, x, y, w, h, PATCOPY);
    TextOut(hdc, x + MERGIN_X, y + MERGIN_Y, timestamp, strlen(timestamp));
}

BOOL CALLBACK
procRecordDialog(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
    case WM_INITDIALOG:
        moveRecordDialog(hWnd);
        break;

    case WM_COMMAND:
        switch (LOWORD(wParam)) {
        case IDC_BT_STOP:
            SendMessage(hWnd, WM_CLOSE, 0, TRUE);
            break;

        case IDC_BT_CANCEL:
            SendMessage(hWnd, WM_CLOSE, 0, FALSE);
            break;
        }
        break;

    case WM_CLOSE:
        closeRecordDialog(hWnd);
        return TRUE;
    }

    return FALSE;
}

BOOL
closeRecordDialog(HWND hWnd)
{
    ::hRecordDialog = NULL;
    ::canContinueRecording = 0;
    updateSettingRecordDialog(hWnd);
    DestroyWindow(hWnd);
    return TRUE;
}

BOOL
moveRecordDialog(HWND hWnd)
{
    RECT rcDlg;
    RECT rcDesktop;

    switch (config.buttonMode) {
    case PREVIOUS_POSITION:
        SetWindowPos(hWnd, NULL, config.buttonPosX, config.buttonPosY, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
        break;

    case CORNER:
        GetWindowRect(hWnd, &rcDlg);
        SystemParametersInfo(SPI_GETWORKAREA, 0, &rcDesktop, 0);
        SetWindowPos(hWnd, NULL, rcDesktop.left,rcDesktop.bottom - (rcDlg.bottom - rcDlg.top),
                     0, 0, SWP_NOSIZE | SWP_NOZORDER);
        break;
    }

    return TRUE;
}

BOOL
updateSettingRecordDialog(HWND hWnd)
{
    if (config.buttonMode == PREVIOUS_POSITION) {
        RECT rc;
        GetWindowRect(hWnd, &rc);
        config.buttonPosX = rc.left;
        config.buttonPosY = rc.top;
        config.updateAll();
    }

    return TRUE;
}

